import os
import csv
import glob
import fnmatch
import numpy as np
import MoNeT_MGDrivE.experimentsHandler as exPar
import MoNeT_MGDrivE.auxiliaryFunctions as auxFun


def writeFactorialAnalysisCSV(
    releasesNumber,
    coverage,
    path,
    experimentString,
    aggregateData,
    ratiosDictionary
):
    """
    Description:
        * Intermediate-level function to write the factorial analysis of the
            factorial experiments to drive.
    In:
        * releasesNumber:
        * coverage:
        * path:
        * experimentString:
        * aggregateData:
        * ratiosDictionary:
    Out:
        * NA
    Notes:
        * NA
    """
    # Getting common data for easier readability
    pop = aggregateData["population"]
    simDays = len(pop)
    numeratorList = ratiosDictionary["numerator"]
    denominatorList = ratiosDictionary["denominator"]
    with open(path+experimentString+".csv", 'wb') as csvfile:
        writer = csv.writer(csvfile, delimiter=',')
        writer.writerow(["ReleasesNumber", "Coverage", "Day", "Ratio"])
        for i in range(0, simDays):
            num = np.sum(pop[i, [numeratorList]])
            denom = np.sum(pop[i, [denominatorList]])
            ratio = num/denom
            writer.writerow([releasesNumber, coverage, i, ratio])


def loadFolderAndWriteFactorialCSV(
    experimentString,
    path,
    aggregationDictionary,
    ratiosDictionary,
    male=True,
    female=True,
    dataType=float
):
    """
    Description:
        * Wrapper function to perform the whole factorial parsing analysis on a
            folder and write the resulting CSV to drive.
    In:
        * experimentString:
        * path: Directory where the experiment is stored.
        * aggregationDictionary: Dictionary containing the keys to aggregate
            the genotypes (created with "generateAggregationDictionary")
        * ratiosDictionary: "numerator", and "denominator" lists dictionary
            containing the columns to use in each section of the ratio.
    Out:
        * NA
    Notes:
        * NA
    """
    filenames = exPar.readExperimentFilenames(path+experimentString)
    landscapeSumData = exPar.sumLandscapePopulationsFromFiles(
        filenames,
        male=True,
        female=True,
        dataType=dataType
        )
    aggregateData = exPar.aggregateGenotypesInNode(
        landscapeSumData,
        aggregationDictionary
    )
    split = auxFun.splitExperimentString(experimentString)
    writeFactorialAnalysisCSV(
        split["releasesNumber"],
        int(split["coverage"])/1000.0,
        path,
        experimentString,
        aggregateData,
        ratiosDictionary
    )
    return None


def compileFactorialCSVFromFiles(path, outFilename):
    """
    Description:
        * Takes all the CSV files from factorial experiments slots
            (starting with 'E_'), and compiles it into a flattened file. The
            file is written directly to directory.
    In:
        * path: Experiment path containing the files.
        * outFilename: (avoid the filename beginning with 'E_').
    Out:
        * NA
    Notes:
        * This function writes to HDD.
    """
    # files = os.listdir(path)
    # filenames = [
    #     path + name for name in files if fnmatch.fnmatch(name, 'E_*.csv')
    # ]
    filenames = glob.glob(path+'E_*.csv')
    f = open(path+outFilename, 'wb')
    for file in filenames:
        fileData = np.genfromtxt(file, skip_header=1, delimiter=",")
        np.savetxt(f, fileData, fmt='%2.5f', delimiter=",")
    f.close()



import os
import csv


def readGenotypes(filename, skipColumns=1):
    '''
    Description:
        * Returns the genotypes header from an run file.
    In:
        * filename: Path to the run file
        * skipColumns: Optional argument for compatibility in the cases where
            the patch number is still present in the data.
    Out:
        * List of genotypes strings
    Notes:
        * NA
    '''
    reader = csv.reader(open(filename))
    return next(reader)[skipColumns:]


def generateAggregationDictionary(genotypesList, indicesList):
    '''
    Description:
        * Creates a dictionary to be used in the aggregation functions.
    In:
        * List of genotypes strings (ordered)
        * List of the indices to aggregate
            (ordered and with duplicate instances)
    Out:
        * Dictionary containing:
            genotypesList
            indices
    Notes:
        * Probably unnecessary, but here so that I don't forget the names
            of the dictionary labels.
    '''
    aggregationDictionary = {
        "genotypes": genotypesList,
        "indices": indicesList
    }
    return aggregationDictionary


def listDirectoriesInPath(path):
    """
    Description:
        * This auxiliary function returns the names of all the folders within
            a given path.
    In:
        * path: Directory address
    Out:
        * List of folder names
    Notes:
        * NA
    """
    file = os.listdir(path)
    folderNames = [
        name for name in file if os.path.isdir(os.path.join(path, name))
    ]
    return folderNames


def splitExperimentString(experimentString):
    """
    Description:
        * Ad hoc auxiliary function to retrieve releases number and coverage
            from the folder name.
    In:
        * Experiment folder name.
    Out:
        * Dictionary containing:
            releasesNumber
            coverage
    Notes:
        * Not to be used in different experiments
    """
    split = experimentString.split("_")
    releasesNumber = int(split[3])
    coverage = int(split[4])
    return {"releasesNumber": releasesNumber, "coverage": coverage}


def findGeneInGenotypesList(gene, genotypes):
    '''
    Description:
        * Returns a list of the repetitions of the allele in a list of
            genotypes strings
    In:
        * gene: character of the allele we are looking for
        * genotypes: list of the genotypes present in the experiment
    Out:
        * List containing the counts of the appearances of the gene in
            the genotypes list
    Notes:
        * This function is mostly designed to be used in a wrapper
            (autoGenerateGenotypesDictionary)
    '''
    repsList = []
    for i in range(0, len(genotypes)):
        count = genotypes[i].count(gene)
        if count > 0:
            repsList.append([i] * count)
    return flatten(repsList)


def flatten(lst):
    '''
    Description:
        * Recursive function to flatten a list to a level 1
    In:
        * Any list
    Out:
        * Flattened list
    Notes:
        * Not coded by me
    '''
    return sum(([x] if not isinstance(x, list) else flatten(x)
                for x in lst), [])


def getUniqueGenesFromGenotypes(genotypes):
    return list({l for word in genotypes for l in word})


def autoGenerateGenotypesDictionary(genes, genotypes):
    '''
    Description:
        * This function takes the genes and genotypes lists, and auto-generates
            the dictionary for the aggregation of genes.
    In:
        * genes: list of the genes to aggregate
        * genotypes: list of strings containing the genotypes
    Out:
        * Aggregation dictionary
    Notes:
        * This is the most general case of the aggregation. For specific
            applications, other functions will need to be created.
    '''
    tempList = []
    for gene in genes:
        tempList.append(findGeneInGenotypesList(gene, genotypes))
    returnDict = {
        "genotypes": genes,
        "indices": tempList
    }
    return returnDict


def listDirectoriesWithPathWithinAPath(pathFilename):
    '''
    Description:
        * Returns a list of folder paths within a given path.
    In:
        * pathFilename: Path to search directories in
    Out:
        * List of folder paths
    Notes:
        * NA
    '''
    folders = listDirectoriesInPath(pathFilename)
    for i, folder in enumerate(folders):
        folders[i] = pathFilename + folder
    return folders



import glob
import numpy as np
import warnings as warnings
import MoNeT_MGDrivE.auxiliaryFunctions as auxFun


def readExperimentFilenames(
    experimentPath,
    sexFilenameIdentifiers={"male": "ADM", "female": "AF1"}
):
    """
    Description:
        * This auxiliary function searches within a given path, and returns all
            the CSV files that match the head provided by the dictionary keys
            for male and female names.
    In:
        * experimentPath: Path to the directory that contains the experiments'
            CSV files (quantiles over patches).
        * sexFilenameIdentifiers: Dictionary containing the head identifier for
            the male, and female CSV files.
    Out:
        * Dictionary with the sorted lists of male and female lexicographically
            sorted patches filenames:
            "male" [list -> strings]
            "female" [list -> strings]
    Notes:
        * This function assumes that male and female naming conventions keep
            the same lexicographical sorting for processing in further
            functions on the pipeline.
    """

    maleFiles = sorted(
        glob.glob(
            experimentPath + "/" +
            sexFilenameIdentifiers["male"] + "*.csv"
        )
    ) if "male" in sexFilenameIdentifiers else []

    femaleFiles = sorted(
        glob.glob(
            experimentPath + "/" +
            sexFilenameIdentifiers["female"] + "*.csv"
        )
    ) if "female" in sexFilenameIdentifiers else []

    return {"male": maleFiles, "female": femaleFiles}


def loadNodeData(
    maleFilename=None,
    femaleFilename=None,
    dataType=float,
    skipHeader=1,
    skipColumns=1
):
    """
    Description:
        * Loads the data for a single node in the files. If male and female
            filenames are provided, it sums them as a matrix operation.
    In:
        * maleFilename: Path to the male CSV file to process.
        * femaleFilename: Path to the female CSV file to process.
        * dataType: To save memory/processing time if possible (int/float).
    Out:
        * Dictionary containing:
            "genotypes" [list -> strings]
            "population" [numpyArray]
    Notes:
        * Timing information is dropped.
    """
    if (maleFilename is not None) and (femaleFilename is not None):
        genotypes = auxFun.readGenotypes(maleFilename)
        dataM = np.genfromtxt(
            maleFilename,
            dtype=dataType,
            skip_header=skipHeader,
            delimiter=","
        )
        dataF = np.genfromtxt(
            femaleFilename,
            dtype=dataType,
            skip_header=skipHeader,
            delimiter=","
        )
        returnDictionary = {
            "genotypes": genotypes,
            "population": (dataM + dataF)[:, skipColumns:]
        }
        return returnDictionary
    elif femaleFilename is not None:
        genotypes = auxFun.readGenotypes(femaleFilename)
        dataF = np.genfromtxt(
            femaleFilename,
            dtype=dataType,
            skip_header=skipHeader,
            delimiter=","
        )
        returnDictionary = {
            "genotypes": genotypes,
            "population": dataF[:, skipColumns:]
        }
        return returnDictionary
    elif maleFilename is not None:
        genotypes = auxFun.readGenotypes(maleFilename)
        dataM = np.genfromtxt(
            maleFilename,
            dtype=dataType,
            skip_header=skipHeader,
            delimiter=","
        )
        returnDictionary = {
            "genotypes": genotypes,
            "population": dataM[:, skipColumns:]
        }
        return returnDictionary
    else:
        warnings.warn(
            '''No data was loaded because both male and female filenames
            are: None''',
            Warning
        )
        return None


def sumLandscapePopulationsFromFiles(
    filenames,
    male=True,
    female=True,
    dataType=float,
    skipHeader=1,
    skipColumns=1
):
    """
    Description:
        * Use this function if only interested in the population dynamics
            without the spatial component.
        * Sums the population dynamics data across nodes 'in site' for memory
            saving purposes.
    In:
        * filenames: Dictionary with male/female filenames.
        * male: Boolean to select male files for the aggregation.
        * female: Boolean to select female files for the aggregation.
        * dataType: Data type to save memory/processing time if possible.
    Out:
        * Dictionary containing:
            "genotypes" [list -> strings]
            "population" [numpyArray]
    Notes:
        * NA
    """
    # Store the lengths of the filenames lists for error checking
    maleFilesNumber = len(filenames["male"])
    femaleFilesNumber = len(filenames["female"])
    maleFilesEqualFemales = (maleFilesNumber == femaleFilesNumber)
    filesExist = (maleFilesNumber >= 1 and femaleFilesNumber >= 1)
    # Select the appropriate aggregation scheme: male+female, male, female
    if (male and female) and filesExist and maleFilesEqualFemales:
        placeholder = loadNodeData(
            filenames["male"][0],
            filenames["female"][0],
            dataType=dataType,
            skipHeader=skipHeader,
            skipColumns=skipColumns
        )
        genotypes = placeholder["genotypes"]
        tempAggregation = placeholder["population"]
        for i in range(1, maleFilesNumber):
            tempAggregation = tempAggregation + loadNodeData(
                filenames["male"][i],
                filenames["female"][i],
                dataType=dataType,
                skipHeader=skipHeader,
                skipColumns=skipColumns
            )["population"]
        returnDictionary = {
            "genotypes": genotypes,
            "population": tempAggregation
        }
        return returnDictionary
    elif female and femaleFilesNumber >= 1:
        placeholder = loadNodeData(
            None,
            filenames["female"][0],
            dataType=dataType,
            skipHeader=skipHeader,
            skipColumns=skipColumns
        )
        genotypes = placeholder["genotypes"]
        tempAggregation = placeholder["population"]
        for i in range(1, femaleFilesNumber):
            tempAggregation = tempAggregation + loadNodeData(
                None,
                filenames["female"][i],
                dataType=dataType,
                skipHeader=skipHeader,
                skipColumns=skipColumns
            )["population"]
        returnDictionary = {
            "genotypes": genotypes,
            "population": tempAggregation
        }
        return returnDictionary
    elif male and maleFilesNumber >= 1:
        placeholder = loadNodeData(
            filenames["male"][0],
            None,
            dataType=dataType,
            skipHeader=skipHeader,
            skipColumns=skipColumns
        )
        genotypes = placeholder["genotypes"]
        tempAggregation = placeholder["population"]
        for i in range(1, maleFilesNumber):
            tempAggregation = tempAggregation + loadNodeData(
                filenames["male"][i],
                None,
                dataType=dataType,
                skipHeader=skipHeader,
                skipColumns=skipColumns
            )["population"]
        returnDictionary = {
            "genotypes": genotypes,
            "population": tempAggregation
        }
        return returnDictionary
    else:
        warnings.warn(
            '''No data was loaded. Check that at least one of the sexes is
            selected, and that the filenames list is not empty.''',
            Warning
        )
        return None


def aggregateGenotypesInNode(
    nodeData,
    aggregationDictionary
):
    """
    Description:
        * This function sums the columns in the way described by the provided
            list, so that we can analyze alleles counts in the populations.
    In:
        * nodeData: Data loaded with the "loadNodeData" function.
        * aggregationDictionary: Genotypes and indices counts dictionary.
    Out:
        * Dictionary containing:
            "genotypes" [list -> strings]
            "population" [numpyArray]
    Notes:
        * A wrapper function should be created to select the columns
            automatically.
    """
    nodeData = nodeData["population"]
    shape = nodeData.shape
    finalGenotypesNumber = len(aggregationDictionary["genotypes"])
    genotypesIndices = aggregationDictionary["indices"]
    fillArray = np.empty([shape[0], finalGenotypesNumber], dtype=float)
    for i in range(0, finalGenotypesNumber):
        cols = genotypesIndices[i]
        fillArray[:, i] = np.sum(nodeData[:, cols], axis=1)
    returnDictionary = {
        "genotypes": aggregationDictionary["genotypes"],
        "population": fillArray
    }
    return returnDictionary


def loadLandscapeData(filenames, male=True, female=True, dataType=float):
    """
    Description:
        * Imports the information of all the nodes within an experiment's
            folder.
    In:
        * filenames: Dictionary with male/female filenames.
    Out:
        * Dictionary containing:
            "genotypes" [list -> strings]
            "landscape" [list -> numpyArrays]
    Notes:
        * NA
    """
    maleFilesNumber = len(filenames["male"])
    femaleFilesNumber = len(filenames["female"])
    maleFilesEqualFemales = (maleFilesNumber == femaleFilesNumber)
    filesExist = (maleFilesNumber >= 1 and femaleFilesNumber >= 1)
    # Select the appropriate aggregation scheme: male+female, male, female
    if (male and female) and filesExist and maleFilesEqualFemales:
        maleFilenames = filenames["male"]
        femaleFilenames = filenames["female"]
        genotypes = auxFun.readGenotypes(maleFilenames[0])
        nodesDataList = [None]*maleFilesNumber
        for i in range(0, maleFilesNumber):
            nodesDataList[i] = loadNodeData(
                maleFilenames[i],
                femaleFilenames[i],
                dataType=dataType
            )["population"]
        returnDictionary = {
            "genotypes": genotypes,
            "landscape": nodesDataList
        }
        return returnDictionary
    elif female and femaleFilesNumber >= 1:
        femaleFilenames = filenames["female"]
        genotypes = auxFun.readGenotypes(femaleFilenames[0])
        nodesDataList = [None] * femaleFilesNumber
        for i in range(0, femaleFilesNumber):
            nodesDataList[i] = loadNodeData(
                None,
                femaleFilenames[i],
                dataType=dataType
            )["population"]
        returnDictionary = {
            "genotypes": genotypes,
            "landscape": nodesDataList
        }
        return returnDictionary
    elif male and maleFilesNumber >= 1:
        maleFilenames = filenames["male"]
        genotypes = auxFun.readGenotypes(maleFilenames[0])
        nodesDataList = [None] * maleFilesNumber
        for i in range(0, maleFilesNumber):
            nodesDataList[i] = loadNodeData(
                maleFilenames[i],
                None,
                dataType=dataType
            )["population"]
        returnDictionary = {
            "genotypes": genotypes,
            "landscape": nodesDataList
        }
        return returnDictionary
    else:
        warnings.warn(
            '''No data was loaded because both male and female filenames
                are: None; or there are no files to load.''',
            Warning
        )
        return None


def aggregateGenotypesInLandscape(landscapeData, aggregationDictionary):
    """
    Description:
        * Aggregates the genotype counts in a landscape according to the
            indices described in the aggregation dictionary.
    In:
        * landscapeData: Data loaded with the "loadLandscapeData" function.
        * aggregationDictionary: Genotypes and indices counts dictionary.
    Out:
        * Dictionary containing:
            "genotypes" [list -> strings]
            "landscape" [list -> numpyArrays]
    Notes:
        * NA
    """
    length = len(landscapeData["landscape"])
    tempList = [None]*length
    for i in range(0, length):
        tempList[i] = aggregateGenotypesInNode(
            {
                "genotypes": landscapeData["genotypes"],
                "population": landscapeData["landscape"][i]
            },
            aggregationDictionary
        )["population"]
    returnDictionary = {
        "genotypes": aggregationDictionary["genotypes"],
        "landscape": tempList
    }
    return returnDictionary


def loadAndAggregateLandscapeData(
    filenames,
    aggregationDictionary,
    male=True,
    female=True,
    dataType=float
):
    """
    Description:
        * Loads and aggregates the data from a landscape given an aggregation
            dictionary (atomic version of separate operations in previous
            versions)
    In:
        * filenames:
        * aggregationDictionary:
        * male: Boolean to select male files for the aggregation.
        * female: Boolean to select female files for the aggregation.
        * dataType: Data type to save memory/processing time if possible.
    Out:
        * aggLandscape: Aggregated landscape by genotypes.
    Notes:
        * This does not sum the nodes, so the return structure is still an
            array of arrays with the nodes' information (within a dictionary).
    """
    rawLandscape = loadLandscapeData(
        filenames, male=male, female=female, dataType=dataType
    )
    aggLandscape = aggregateGenotypesInLandscape(
        rawLandscape,
        aggregationDictionary
    )
    return aggLandscape


def loadAndAggregateLandscapeDataRepetitions(
    paths,
    aggregationDictionary,
    male=True,
    female=True,
    dataType=float
):
    """
    Description:
        * Loads and aggregates the genotypes of the landscape accross
            repetitions of the same experiment.
    In:
        * paths: Repetitions folders locations.
        * aggregationDictionary: Genotypes and indices counts dictionary.
        * male: Boolean to select male files for the aggregation.
        * female: Boolean to select female files for the aggregation.
        * dataType: Data type to save memory/processing time if possible.
    Out:
        * returnDict: Dictionary with genotypes and the loaded landscapes
            for each one of the repetitions.
    Notes:
        * This function is meant to work with the traces plot, so it has a
            higher dimension (repetitions) than regular spatial analysis
            versions.
    """
    pathsNumber = len(paths)
    landscapes = [None] * pathsNumber
    for i in range(0, pathsNumber):
        filenames = readExperimentFilenames(paths[i])
        loadedLandscape = loadAndAggregateLandscapeData(
            filenames, aggregationDictionary,
            male=male, female=female, dataType=dataType
        )
        landscapes[i] = loadedLandscape["landscape"]
    returnDict = {
        "genotypes": aggregationDictionary["genotypes"],
        "landscapes": landscapes
    }
    return returnDict


def getGenotypeFromLandscape(
    landscapeData,
    genotypeIndex
):
    """
    Description:
        * Returns the spatiotemporal array of the genotype queried
            (through index)
    In:
        * landscapeData: Aggregated landscape dictionary.
        * genotypeIndex: Integer location of the genotype needed.
    Out:
        * geneArray: Array with the gene amounts per node (through time).
    Notes:
        * NA
    """
    nodesNumb = len(landscapeData["landscape"])
    time = len(landscapeData["landscape"][0])
    geneArray = np.empty([nodesNumb, time])
    for i in range(0, nodesNumb):
        probe = landscapeData["landscape"][i]
        geneArray[i] = [row[genotypeIndex] for row in probe]
    return geneArray


def getGenotypeArraysFromLandscape(
    landscapeData
):
    """
    Description:
        * Returns the spatiotemporal arrays of all the genotypes present in the
            landscape
    In:
        * landscapeData: population dynamics data
    Out:
        * geospatialDict: contains the genotypes and spatiotemporal
            gene landscape
    Notes:
        * NA
    """
    genotypes = landscapeData["genotypes"]
    genesSpatialList = [None] * len(genotypes)
    for i in range(0, len(genotypes)):
        genesSpatialList[i] = getGenotypeFromLandscape(landscapeData, i)
    geospatialDict = {
        "genotypes": genotypes,
        "geneLandscape": genesSpatialList
    }
    return geospatialDict



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
	  name='MoNeT_MGDrivE',
      version='0.3.5',
      url='https://chipdelmal.github.io/MoNeT/',
      author='Hector M. Sanchez C.',
      author_email='sanchez.hmsc@berkeley.edu',
      description="MoNeT python package",
      long_description=long_description,
      long_description_content_type="text/markdown",
      packages=setuptools.find_packages(),
      install_requires=['numpy','scipy','matplotlib','ipython','jupyter','pandas','sympy'],
      license='MIT',
      classifiers=[
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
      ],
 )



import numpy as np


def deHashFactorial(hashedResults):
    '''
    Description:
        * Converts a factorial experiment back from a dictionary form into a
            x,y,z array.
    In:
        * hashedResults: Dictionary containing the keys and values of the
            experiment.
    Out:
        * Array containing the slots of the factorial experiment.
    Notes:
        * Only works for 3d factorial experiments.
    '''
    hashLength = len(hashedResults)
    returnArray = np.empty([hashLength, 3])
    for (i, key) in enumerate(hashedResults.keys()):
        splits = key.split("-")
        returnArray[i] = [int(splits[0]), float(splits[1]), hashedResults[key]]
    return returnArray


def calculateFactorialHashError(probeData, centralData, errorFunction):
    '''
    Description:
        * Calculates the error between two hashed factorial datasets by
            applying the error function provided.
    In:
        * probeData: Data to be taken to calculate error.
        * centralData: Data to be taken for "truth" in the experiment.
        * errorFunction: 2-input function to calculate the difference between
            datasets.
    Out:
        * Hash containing the application of the error function to each of the
            pairs of entries in the factorial set.
    Notes:
        * Only works for single output factorial data.
    '''
    centralKeys = centralData.keys()
    probeKeys = probeData.keys()
    resultsHash = {}
    for key in probeKeys:
        if(key in centralKeys):
            trueValue = centralData[key]
            probeValue = probeData[key]
            error = errorFunction(probeValue, trueValue)
            resultsHash.update({key: error})
    return resultsHash


def loadAndHashFactorialCSV(
    path,
    probedDay=None,
    floatMultiplier=1,
    skipHeader=0
):
    '''
    Description:
        * Imports a dataset into a dictionary filtered at a single day.
    In:
        * path: Directory where the factorial experiment is stored.
        * probedDay: Day at which the data should be probed.
        * floatMultiplier: DEPRECATED
    Out:
        * Dictionary containing the entries on the selected day.
    Notes:
        * If no day is provided, the last day of the simulation is returned.
    '''
    factorialData = np.genfromtxt(
        path,
        dtype=float,
        skip_header=skipHeader,
        delimiter=","
    )
    if(probedDay is not None and probedDay >= 1):
        dayProbe = probedDay
    else:
        dayProbe = max(factorialData[:, 2])
    probedHash = prepareFactorialArrayForComparison(
        factorialData,
        dayProbe,
        floatMultiplier
    )
    return probedHash


def prepareFactorialArrayForComparison(
    factorialData,
    probeDay,
    floatMultiplier=1
):
    '''
    Description:
        * Wrapper function that takes the array of factorial data, and returns
            the hashed version.
    In:
        * factorialData: Loaded data to convert into a dictionary.
        * probeDay: Day at which the data will be filtered.
        * floatMultiplier: Re-scales the coverage level to make
            integer instead of float (important for hashing differences).
    Out:
        * Dictionary containing the hashed factorial experiment.
    Notes:
        *
    '''
    probedData = getReleasesCoverageRatioArray(factorialData, probeDay)
    probedHashMap = convertFactorialArrayToHash(probedData, floatMultiplier)
    return probedHashMap


def convertFactorialArrayToHash(factorialArray, floatMultiplier=1):
    '''
    Description:
        * Low level fuction to convert an array into a factorial experiment
            dictionary
    In:
        * factorialArray: Array containing the factorial slots.
        * floatMultiplier: Multiple of 10 to transform the float into an
            integer if needed.
    Out:
        * Dictionary containing the factorial experiment.
    Notes:
        * Ad-hoc function for the threshold-dependent paper. It assumes that
            the first quantity is an integer, and the second one is a float.
            The float precision is limited to 5 figures.
    '''
    returnDictionary = {}
    for i in range(0, len(factorialArray)):
        releases = str(int(factorialArray[i, 0])).rjust(3, "0")
        coverage = str(
            float(factorialArray[i, 1] * floatMultiplier)
        ).ljust(5, "0")
        key = releases + "-" + coverage
        returnDictionary.update({key: factorialArray[i, 2]})
    return returnDictionary


def getReleasesCoverageRatioArray(factorialArray, probeDay):
    '''
    Description:
        * Ad hoc function to retrieve the {releases, coverage, ratio} triplet
            from the {releases, coverage, day, ratio} at a given day.
    In:
        * factorialArray: raw factorial data containing the days in the
            columns.
        * probeDay: day to pick the information out of.
    Out:
        * factorialPoints: array containing the factorial points of the
            experiment.
    Notes:
        * NA
    '''
    probedData = factorialArray[factorialArray[:, 2] == probeDay]
    factorialPoints = probedData[:, [0, 1, 3]]
    return factorialPoints


def sampleDifference(x, y): return (x - y)


def sampleError(test, real): return (real - test) / real
